/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * backward computation for math operations
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-18
 */

#include "XNoder.h"
#include "XBackwardMath.h"
#include "../tensor/XName.h"
#include "../tensor/core/CHeader.h"

namespace nts{

/* compute dE/dx of a node */
void XMathGrad::MakeGrad(XTensor * node)
{
    CheckNTErrors(node->grad != NULL, "No gradient found!");

    XLink &income = node->income;
    int operID = income.typeID;

    if(operID == MATH_SUM)
        GradSum(node);
    else if(operID == MATH_SUMDIM)
        GradSumDim(node);
    else if(operID == MATH_MULTIPLY)
        GradMultiply(node);
    else if(operID == MATH_MATRIXMUL)
        GradMatrixMul(node);
    else{
        ShowNTErrors("TODO!");
    }
}

/* indicates whether the node is for a math operation */
bool XMathGrad::IsMathOP(XTensor * node)
{
    XLink &income = node->income;
    return (income.typeID & MATH_BASE) != 0;
}

/* 
gradient for sum
for 
c =  a + b * \beta
we have
dE/da = dE/dc 
dE/db = dE/dc * \beta
>> node - the node (c) for backward computation
*/
void XMathGrad::GradSum(XTensor * node)
{
    XLink &income = node->income;
    CheckNTErrors(income.tailNum == 2, "Wrong input tensor number for SUM!");

    XTensor * a = income.tails[0];
    XTensor * b = income.tails[1];
    DTYPE beta = income.GetParam(0);
    XNoder::MakeGrad(a);
    XNoder::MakeGrad(b);

    _Sum(a->grad, node->grad, a->grad);
    _Sum(b->grad, node->grad, b->grad, beta);

    node->visitMark = NODE_FINISHED;
}

/* 
gradient for sum with one dimension
c = a + b * \beta
where the size of b is equal to dimension n of a, i.e., |b| = a.dimSize[n]
dE/da = dE/dc
dE/db = dE/dc * b.reduce(0,...,n-1,n+1,...) * \beta
*/
void XMathGrad::GradSumDim(XTensor * node)
{
    XLink &income = node->income;
    CheckNTErrors(income.tailNum == 2, "Wrong input tensor number for SUMDIM!");

    XTensor * a = income.tails[0];
    XTensor * b = income.tails[1];
    int n = income.GetParamInt(0);
    DTYPE beta = income.GetParam(1);
    XNoder::MakeGrad(a);
    XNoder::MakeGrad(b);

    _Sum(a->grad, node->grad, a->grad);

    int order = a->order;
    int dimSize[MAX_TENSOR_DIM_NUM];
    memcpy(dimSize, a->dimSize, sizeof(int) * a->order);

    if(n == order - 1){
        int reshapedSize[MAX_TENSOR_DIM_NUM];
        reshapedSize[0] = a->unitNum/dimSize[order - 1];
        reshapedSize[1] = dimSize[order - 1];

        /* we reshape dE/dc to a matrix whose column number is equal to the 
           size of b. Then we can reduce the matrix into a row vector. */
        node->grad->Reshape(2, reshapedSize);

        if(b->outgo.tailNum > 1){
            XTensor * bGradTMP = NewTensorBuf(b->grad, b->devID, b->mem);
            _ReduceSum(node->grad, bGradTMP, 0);
            if(beta != 1.0F)
                _ScaleAndShiftMe(bGradTMP, beta);
            _Sum(bGradTMP, b->grad, b->grad);
            DelTensorBuf(bGradTMP);
        }
        else{
            _ReduceSum(node->grad, b->grad, 0);
            if(beta != 1.0F)
                _ScaleAndShiftMe(b->grad, beta);
        }

        node->grad->Reshape(order, dimSize);
    }
    else{
        int reshapedSize[MAX_TENSOR_DIM_NUM];
        reshapedSize[0] = 1;
        reshapedSize[1] = dimSize[n];
        reshapedSize[2] = 1;

        for(int i = 0; i < order; i++){
            if(i < n)
                reshapedSize[0] *= dimSize[i];
        }

        reshapedSize[2] = a->unitNum / (reshapedSize[0] * reshapedSize[1]);

        /* we reshape dE/dc to a 3D tensor of size (x, y, z) where y = |b|. 
           Then reduce along with z and x to obtain dE/db. */
        node->grad->Reshape(3, reshapedSize);

        XTensor * interGrad = NewTensorBuf(2, reshapedSize, b->dataType, b->denseRatio, b->devID, b->mem);

        _ReduceSum(node->grad, interGrad, 2);

        if(b->outgo.tailNum > 1){
            XTensor * bGradTMP = NewTensorBuf(b->grad, b->devID, b->mem);
            _ReduceSum(interGrad, bGradTMP, 0);
            if(beta != 1.0F)
                _ScaleAndShiftMe(bGradTMP, beta);
            _Sum(bGradTMP, b->grad, b->grad);
            DelTensorBuf(bGradTMP);
        }
        else{
            _ReduceSum(interGrad, b->grad, 0);
            if(beta != 1.0F)
                _ScaleAndShiftMe(b->grad, beta);
        }

        node->grad->Reshape(order, dimSize);

        DelTensorBuf(interGrad);

    }
    
    node->visitMark = NODE_FINISHED;
}

/* 
gradient for multiply (dot production)
for
c =  a * b 
we have
dE/da = dE/dc * b
dE/db = dE/dc * a 
>> node - the node (c) for backward computation
*/
void XMathGrad::GradMultiply(XTensor * node)
{
    XLink &income = node->income;
    CheckNTErrors(income.tailNum == 2, "Wrong input tensor number for MULTIPLY!");

    XTensor * a = income.tails[0]; 
    XTensor * b = income.tails[1];
    XNoder::MakeGrad(a);
    XNoder::MakeGrad(b);

    CheckNTErrors(XTensor::IsSameShaped(a, b), "Wrong sized input tensors!");
    _Multiply(node->grad, b, a->grad, 1.0F);
    _Multiply(node->grad, a, b->grad, 1.0F);

    node->visitMark = NODE_FINISHED;
}

/* 
gradient for matrix multiply
for c = matmul(a, b) * \alpha
we have 
dE/da = dE/dc * b^T * \alpha
dE/db = a^T * dE/dc * \alpha
>> node - the node (c) for backward computation
*/
void XMathGrad::GradMatrixMul(XTensor * node)
{
    XLink &income = node->income;
    CheckNTErrors(income.tailNum == 2, "Wrong input tensor number for MULTIPLY!");
    CheckNTErrors(income.paramNum == 3, "Wrong parameter number for MULTIPLY!");

    XTensor * a = income.tails[0]; 
    XTensor * b = income.tails[1];
    MATRIX_TRANS_TYPE transA = income.GetParamTrans(0);
    MATRIX_TRANS_TYPE transB = income.GetParamTrans(1);
    DTYPE alpha = income.GetParam(2);

    XNoder::MakeGrad(a);
    XNoder::MakeGrad(b);

    XTensor * dedc = node->grad;
    XTensor * deda = a->grad;
    XTensor * dedb = b->grad;
    
    /* c = a * b * \alpha */
    if(transA == X_NOTRANS && transB == X_NOTRANS){

        /* dE/da = dE/dc * b^T * \alpha */
        _MatrixMul(dedc, X_NOTRANS, b, X_TRANS, deda, alpha, 1.0F);

        /* dE/db = a^T * dE/dc * \alpha */
        _MatrixMul(a, X_TRANS, dedc, X_NOTRANS, dedb, alpha, 1.0F);
    }
    
    /* c = a^T * b * \alpha */
    else if(transA == X_TRANS && transB == X_NOTRANS){
        
        /* dE/da = dE/dc * b^T * \alpha */
        _MatrixMul(dedc, X_NOTRANS, b, X_TRANS, deda, alpha, 1.0F);
        
        /* dE/db = a * dE/dc * \alpha */
        _MatrixMul(a, X_NOTRANS, dedc, X_NOTRANS, dedb, alpha, 1.0F);
    }
    
    /* c = a * b^T * \alpha */
    else if(transA == X_NOTRANS && transB == X_TRANS){
        
        /* dE/da = dE/dc * b * \alpha */
        _MatrixMul(dedc, X_NOTRANS, b, X_NOTRANS, deda, alpha, 1.0F);
        
        /* dE/db = a^T * dE/dc * \alpha */
        _MatrixMul(a, X_TRANS, dedc, X_NOTRANS, dedb, alpha, 1.0F);
    }
    
    /* c = a^T * b^T * \alpha */
    else if(transA == X_TRANS && transB == X_TRANS){
        
        /* dE/da = dE/dc * b * \alpha */
        _MatrixMul(dedc, X_NOTRANS, b, X_NOTRANS, deda, alpha, 1.0F);
        
        /* dE/db = a * dE/dc * \alpha */
        _MatrixMul(a, X_NOTRANS, dedc, X_NOTRANS, dedb, alpha, 1.0F);
    }

    node->visitMark = NODE_FINISHED;
}

}
