/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University. 
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-31
 */

#include <math.h>
#include "T2TEncoder.h"
#include "T2TLayerNormal.h"
#include "T2TUtility.h"
#include "../../tensor/core/CHeader.h"

namespace transformer
{

/* constructor */
AttEncoder::AttEncoder()
{
}

/* de-constructor */
AttEncoder::~AttEncoder()
{
    delete[] attentions;
    delete[] fnns;
    delete[] layerNorms;
}

/* 
initialize the model 
>> argc - number of arguments
>> argv - list pf pointers to the arguments
>> myDevID - device id
>> myMem - the memory pool
*/
void AttEncoder::InitModel(int argc, const char ** argv, int myDevID, XMem * myMem)
{
    devID = myDevID;
    mem = myMem;
    
    LoadParamInt(argc, argv, "nstack", &nlayer, 6);
    LoadParamInt(argc, argv, "hsize", &hSize, 512);
    LoadParamInt(argc, argv, "esize", &eSize, 512);
    LoadParamInt(argc, argv, "vsize", &vSize, -1);


    CheckNTErrors(nlayer > 1, "We have one encoding layer at least!");
    CheckNTErrors(vSize > 1, "set vocabulary size by \"-vsize\"");

    /* embedding model */
    embedder.InitModel(argc, argv, devID, mem);

    attentions = new T2TAttention[nlayer];
    fnns = new T2TFNN[nlayer];
    layerNorms = new T2TLN[nlayer];

    /* initialize the stacked layers */
    for(int i = 0; i < nlayer; i++){
        attentions[i].InitModel(argc, argv, myDevID, myMem);
        fnns[i].InitModel(argc, argv, myDevID, myMem);
        layerNorms[i].InitModel(argc, argv, myDevID, myMem);
    }
}

/* 
make the encoding network
>> input - the input tensor of the encoder
<< return - the output tensor of the encoder
*/
XTensor * AttEncoder::Make(XTensor * input)
{
    XTensor * x = embedder.Make(input);

    for(int i = 0; i < nlayer; i++){
        XTensor * att;
        XTensor * ln;
        XTensor * fnn;
        XTensor res;

        /* self attention */
        att = attentions[i].Make(x, x, x);

        /* residual connection */
        res = Sum(*att, *x);

        /* TODO: dropout */

        /* layer normalization */
        ln = layerNorms[i].Make(&res);

        /* input of next layer */
        x = ln;

        /* fnn */
        fnn = fnns[i].Make(x);

        /* residual connection */
        res = Sum(*fnn, *x);

        /* TODO: dropout */

        /* layer normalization */
        ln = layerNorms[i].Make(&res);

        /* input of next layer */
        x = ln;
    }

    return x;
}

}
